""".. Ignore pydocstyle D400.

=======
Resolwe
=======

Open source enterprise dataflow engine in Django.

"""
from resolwe.__about__ import (  # noqa: F401
    __author__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __url__,
    __version__,
)
